/*
**  SLOPE_FOR: C-XSC classes for one- and multidimensional (interval)
**             slope arithmetic in forward mode
**
**  Copyright (C) 1999-2001 M. Braeuer, W. Hofschuster, W. Kraemer
**                          Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//////////////////////////////////////////////////////////////
//
//Implementierung  multi_dim_slope.cpp
//
//////////////////////////////////////////////////////////////

#include "multi_dim_slope.h"

///////////////////////////////////////////////////////////////
//
//Implementierung der Klasse slope
//
///////////////////////////////////////////////////////////////

namespace cxsc {

//statische Komponenten
int slope::calc_slope=1; 

//------------------------------------------------------------------

void slope::set_calc_slope()
{
 slope::calc_slope=1;
}

//------------------------------------------------------------------

void slope::unset_calc_slope()
{
 slope::calc_slope=0;
}

//------------------------------------------------------------------

int slope::get_calc_slope()
{
 return slope::calc_slope;
}

//------------------------------------------------------------------

//Konstruktoren
slope::slope()
{
 dim=1;
 Resize(f_s,1,dim);
} 

//------------------------------------------------------------------

slope::slope(const int dimension)
{
 dim=dimension;
 Resize(f_s,1,dim);
} 

//------------------------------------------------------------------

slope::slope(const slope& s)
{
 dim=s.dim;
 f_x=s.f_x;
 f_c=s.f_c;
 Resize(f_s,1,dim);
 for (int i=1; i<=dim; i++) f_s[i]=s.f_s[i];
}

//------------------------------------------------------------------

//Zuweisungsoperator
slope& slope::operator=(const slope& s)
{
 if(this!= &s)
   {
     dim=s.dim;
     f_x=s.f_x;
     f_c=s.f_c;
     Resize(f_s,1,dim);
     f_s=s.f_s;
   }
 return *this;  
}

//------------------------------------------------------------------

//Verschiedene Initialisierungsfunktionen

//------------------------------------------------------------------

slope var_slope(int dimension, int nr,const  real& x)
{
 slope s(dimension);
 int i;

 s.f_x=_interval(x);
 s.f_c=_interval(x);
 for(i=1; i<nr; i++) s.f_s[i]=_interval(0.0);
 s.f_s[nr]=_interval(1.0);
 for(i=nr+1; i<=s.dim; i++) s.f_s[i]=_interval(0.0);
 return s;
}

//------------------------------------------------------------------

slope var_slope(int dimension, int nr,const  interval& x)
{
 slope s(dimension);
 int i;

 s.f_x=x;
 s.f_c=x;
 for(i=1; i<nr; i++) s.f_s[i]=_interval(0.0);
 s.f_s[nr]=_interval(1.0);
 for(i=nr+1; i<=s.dim; i++) s.f_s[i]=_interval(0.0);
 return s;
}

//------------------------------------------------------------------

slope var_slope(int dimension, int nr,const  interval& x,const  interval& c)
{
 slope s(dimension);
 int i;

 s.f_x=x;
 s.f_c=c;
 for(i=1; i<nr; i++) s.f_s[i]=_interval(0.0);
 s.f_s[nr]=_interval(1.0);
 for(i=nr+1; i<=s.dim; i++) s.f_s[i]=_interval(0.0);
 return s;
}

//------------------------------------------------------------------

slope var_slope(int dimension, int nr,const  interval& x,const  real& c)
{
 slope s(dimension);
 int i;

 s.f_x=x;
 s.f_c=_interval(c);
 for(i=1; i<nr; i++) s.f_s[i]=_interval(0.0);
 s.f_s[nr]=_interval(1.0);
 for(i=nr+1; i<=s.dim; i++) s.f_s[i]=_interval(0.0);
 return s;
}

//------------------------------------------------------------------

slope const_slope(int dimension,const  real& x)
{
 slope s(dimension);
 int i;

 s.f_x=_interval(x);
 s.f_c=_interval(x);
 for(i=1; i<=s.dim; i++) s.f_s[i]=_interval(0.0);
 return s;
}

//------------------------------------------------------------------

slope const_slope(int dimension,const  interval& x)
{
 slope s(dimension);
 int i;

 s.f_x=x;
 s.f_c=x;
 for(i=1; i<=s.dim; i++) s.f_s[i]=_interval(0.0);
 return s;
}

//------------------------------------------------------------------

//Initialisierung der 2.Intervallkomponente mit Mittelpunkt
//bei der nr-ten Unabhaengigen 
slope init_mid_var(int dimension, int nr,const  interval& x)
{
 slope s(dimension);
 int i;

 s.f_x=x;
 s.f_c=_interval(mid(x));
 for(i=1; i<nr; i++) s.f_s[i]=_interval(0.0);
 s.f_s[nr]=_interval(1.0);
 for(i=nr+1; i<=s.dim; i++) s.f_s[i]=_interval(0.0);
 return s;
}

//------------------------------------------------------------------

//Ausgabe
void print_slope(const slope& s)
{
  cout << "Slope Objekt mit Dimension " << s.dim << endl;

  cout << SaveOpt << SetPrecision(16,14)<< "f_x   : " << s.f_x << endl;
  cout << "f_c   : " << s.f_c << endl;
  
  for(int i=1; i<=s.dim; i++)
  cout << "f_s[" << i << "]: " << s.f_s[i] << endl;
  cout << RestoreOpt << endl; 
}

//------------------------------------------------------------------

//Folgende Funktionen geben Elemente zurueck
int get_dim(const slope& s)
{
 return s.dim;
}

interval get_fx(const slope& s)
{
 return s.f_x;
}

interval get_fc(const slope& s)
{
 return s.f_c;
}

ivector get_fs(const slope& s)
{
 return s.f_s;
}

//------------------------------------------------------------------

//Ueberladen der Operatoren

//------------------------------------------------------------------

//unaerer Operator -
slope operator-(const slope& s)
{
 slope res(s.dim);
 res.f_x=-s.f_x;
 if(slope::calc_slope)
  {
    res.f_c=-s.f_c;
    for(int i=1; i<=s.dim; i++) res.f_s[i]=-s.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

//binaere Operatoren (slope,slope) 
slope operator-(const slope& s,const  slope& t)
{
 int dim1=s.dim;
 int dim2=t.dim;
 if(dim1!=dim2)
   {
    cout << "Fehler in operator-(slope& ,slope& )";
    cout << endl; exit(1);
   }

 slope res(dim1);
 res.f_x=s.f_x-t.f_x;
 if(slope::calc_slope)
  {
    res.f_c=s.f_c-t.f_c;
    for(int i=1; i<=dim1; i++) res.f_s[i]=s.f_s[i]-t.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

slope operator+(const slope& s,const  slope& t)
{
 int dim1=s.dim;
 int dim2=t.dim;
 if(dim1!=dim2)
   {
    cout << "Fehler in operator+(slope& ,slope& )";
    cout << endl; exit(1);
   }

 slope res(dim1);
 res.f_x=s.f_x+t.f_x;
 if(slope::calc_slope)
  {
    res.f_c=s.f_c+t.f_c;
    for(int i=1; i<=dim1; i++) res.f_s[i]=s.f_s[i]+t.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

slope operator*(const slope& s,const  slope& t)
{
 int dim1=s.dim;
 int dim2=t.dim;
 if(dim1!=dim2)
   {
    cout << "Fehler in operator*(slope& ,slope& )";
    cout << endl; exit(1);
   }

 slope res(dim1);
 res.f_x=s.f_x*t.f_x;
 if(slope::calc_slope)
  {
    res.f_c=s.f_c*t.f_c;
    for(int i=1; i<=dim1; i++) 
      {
	res.f_s[i]=t.f_x*s.f_s[i]+s.f_c*t.f_s[i];
      }
  }
 return res;
}

//------------------------------------------------------------------

slope operator/(const slope& s, const slope& t)
{
 int dim1=s.dim;
 int dim2=t.dim;
 if(dim1!=dim2)
   {
    cout << "Fehler in operator/(slope& ,slope& )";
    cout << endl; exit(1);
   }

 slope res(dim1);
 if(_interval(0.0)<=t.f_x) 
   {
     cout << "Fehler in / 0 in f_x" << endl; exit(1);
   }
 res.f_x=s.f_x/t.f_x;
 if(slope::calc_slope)
  {
    if(_interval(0.0)<=t.f_c) 
      {
	cout << "Fehler in / 0 in f_c" << endl; exit(1);
      }
    res.f_c=s.f_c/t.f_c;
    for(int i=1; i<=dim1; i++) 
      {
	res.f_s[i]=(s.f_s[i]-res.f_c*t.f_s[i])/t.f_x;
      }
  }
 return res;
}

//------------------------------------------------------------------

//binaere Operatoren (slope,interval) 
slope operator-(const slope& s,const interval& t)
{
 slope res(s.dim);
 res.f_x=s.f_x-t;
 if(slope::calc_slope)
  {
    res.f_c=s.f_c-t;
    for(int i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

slope operator+(const slope& s, const interval& t)
{
 slope res(s.dim);
 res.f_x=s.f_x+t;
 if(slope::calc_slope)
  {
    res.f_c=s.f_c+t;
    for(int i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

slope operator*(const slope& s, const interval& t)
{
 slope res(s.dim);
 res.f_x=s.f_x*t;
 if(slope::calc_slope)
  {
    res.f_c=s.f_c*t;
    for(int i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]*t;
  }
 return res;
}

//------------------------------------------------------------------

slope operator/(const slope& s, const interval& t)
{
  if(_interval(0.0)<=t) 
   {
     cout << "Fehler in / 0 im Intervall" << endl; exit(1);
   }
 slope res(s.dim);
 res.f_x=s.f_x/t;
 if(slope::calc_slope)
  {
    res.f_c=s.f_c/t;
    for(int i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/t;
  }
 return res;
}

//------------------------------------------------------------------

//binaere Operatoren (interval, slope) 
slope operator-(const interval& s,const  slope& t)
{
 slope res(t.dim);
 res.f_x=s-t.f_x;
 if(slope::calc_slope)
  {
    res.f_c=s-t.f_c;
    for(int i=1; i<=t.dim; i++) res.f_s[i]=-t.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

slope operator+(const interval& s,const  slope& t)
{
 slope res(t.dim);
 res.f_x=s+t.f_x;
 if(slope::calc_slope)
  {
    res.f_c=s+t.f_c;
    for(int i=1; i<=t.dim; i++) res.f_s[i]=t.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

slope operator*(const interval& s,const  slope& t)
{
 slope res(t.dim);
 res.f_x=s*t.f_x;
 if(slope::calc_slope)
  {
    res.f_c=s*t.f_c;
    for(int i=1; i<=t.dim; i++) res.f_s[i]=s*t.f_s[i];
  }
 return res;
}

//------------------------------------------------------------------

slope operator/(const interval& s,const  slope& t)
{
 slope res(t.dim);
 if(_interval(0.0)<=t.f_x) 
   {
     cout << "Fehler in / 0 in f_x" << endl; exit(-1);
   }
 res.f_x=s/t.f_x;
 if(slope::calc_slope)
  {
    if(_interval(0.0)<=t.f_c) 
      {
	cout << "Fehler in / 0 in f_c" << endl; exit(-1);
      }
    res.f_c=s/t.f_c;
    for(int i=1; i<=t.dim; i++) res.f_s[i]=-(res.f_c*t.f_s[i])/t.f_x;
  }
 return res;
}

//------------------------------------------------------------------

//binaere Operatoren (slope,real)
slope operator-( const slope& s, const real& t)
{
 slope res(s.dim);
 res=s-interval(t);
 return res;
}

//------------------------------------------------------------------

slope operator+(const slope& s,const real& t)
{
 slope res(s.dim);
 res=s+_interval(t);
 return res;
}

//------------------------------------------------------------------

slope operator*(const slope& s,const real& t)
{
 slope res(s.dim);
 res=s*_interval(t);
 return res;
}

//------------------------------------------------------------------

slope operator/(const slope& s,const real& t)
{
 slope res(s.dim);
 res=s/_interval(t);
 return res;
}

//------------------------------------------------------------------

//binaere Operatoren (real, slope)
slope operator-(const real& s ,const slope& t)
{
 slope res(t.dim);
 res=_interval(s)-t;
 return res; 
}

//------------------------------------------------------------------

slope operator+(const real& s ,const slope& t)
{
 slope res(t.dim);
 res=_interval(s)+t;
 return res; 
}

//------------------------------------------------------------------

slope operator*(const real& s ,const slope& t)
{
 slope res(t.dim);
 res=_interval(s)*t;
 return res; 
}

//------------------------------------------------------------------

slope operator/(const real& s ,const slope& t)
{
 slope res(t.dim);
 res=_interval(s)/t;
 return res; 
}

//------------------------------------------------------------------

//Ueberladen der Standardfunktionen in Analogie zum eindim. Fall
/*------------------------------------------------------------------------

sqr(x) konvex fuer alle x aus R

------------------------------------------------------------------------*/
slope sqr(const slope& s)
{
 slope res(s.dim);
 res.f_x=sqr(s.f_x);
 if(slope::calc_slope)
  {
    res.f_c=sqr(s.f_c);
    interval h=s.f_x+s.f_c;
    for(int i=1; i<=s.dim; i++) res.f_s[i]=h*s.f_s[i];
  }
 return res;
}


/*------------------------------------------------------------------------

sqrt(x) konkav fuer alle x>=0, streng monoton wachsend

------------------------------------------------------------------------*/
slope sqrt(const slope& s)
{
 slope res(s.dim);
 res.f_x=sqrt(s.f_x);
 if(slope::calc_slope)
  {
    res.f_c=sqrt(s.f_c);
    interval h=res.f_x+res.f_c;
    if(0.0<=h) //0 im Intervall
      {
       cout << "Fehler sqrt: 0 im Intervall" << endl; 
       exit(1); //Funktion ist nicht auswertbar
      }
     else 
       {
	 for(int i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/h;
       }
  }
 return res;
}

/*------------------------------------------------------------------------

sqrt(x,n ) konkav fuer alle x>=0 und n aus N, streng monoton wachsend

------------------------------------------------------------------------*/
slope sqrt(const slope& s,int n)
{
 slope res(s.dim);
 int i;

 if(n==0) 
  {
    cout << "Fehler in sqrt(x,n): n=0" << endl;
    exit(1);
  }

 if(n>0)
   {
     if(n==1) return s;
     if(n==2) return sqrt(s);

     //n>=3:
     res.f_x=sqrt(s.f_x, n);

     if(slope::calc_slope)
       {
	 res.f_c=sqrt(s.f_c, n);

	 interval h=s.f_x;

         real h_d, h_u;

	 interval _x=_interval(Inf(s.f_x));
	 interval _c=_interval(Inf(s.f_c));

	 interval h1= _x-_c; 

	 if(!(_interval(0.0)<=h1)) 
	   {     
	     interval _fx=sqrt(_x,n);
	     interval _fc=sqrt(_c,n);
	     interval i_res=(_fx-_fc)/h1;
	     h_u = Sup(i_res); //s_1_up
	 
	     interval x_=_interval(Sup(s.f_x));
	     interval c_=_interval(Sup(s.f_c));

	     interval h2= x_-c_; 

	     if(!(_interval(0.0)<=h2))
	       {
		 interval fx_=sqrt(x_,n);
		 interval fc_=sqrt(c_,n);
		 interval i_res=(fx_-fc_)/h2;
		 h_d = Inf(i_res); //s_2_down


		 interval sl=_interval(h_d,h_u); 
		 for(i=1; i<=s.dim; i++)
		   {
		     res.f_s[i]=s.f_s[i]*sl;
		   }

		 return res;
	       }          
	   }
	 else
	   {
	     if(0.0<=h) //0 im Intervall
	       {
		 cout << "Fehler sqrt(x,n): 0 im Intervall" << endl; 
		 exit(1); //Funktion nicht auswertbar
	       }
	     else
	       {
		 interval der=sqrt(h,n)/(_interval(n)*h);
		 for(i=1; i<=s.dim; i++)
		   {
		     res.f_s[i]=s.f_s[i]*der;
		   }
	       }
	   }
       }
     return res;
   }
 else {res=_interval(1.0)/sqrt(s,-n); return res;}
}

/*------------------------------------------------------------------------

power(x,n )
    n>0 :   n gerade: konvex fuer alle x aus R fallend, wachsend
            n ungerade: konvex fuer x>=0 str. monoton wachsend
	                konkav fuer x<=0 str. monoton wachsend

------------------------------------------------------------------------*/
slope power(const slope& s,int n)
{
 slope res(s.dim);
 int i;

 if(n==0) 
  {
    if(_interval(0.0)<=s.f_x) 
      {
       cout << "Fehler in power(x,n) " << endl;
       exit(1);
      } 
    real c=1.0;
    res=const_slope(s.dim,c);
    return res;
  }

 if(n>0)
   {
    if(n==1) return s;
    if(n==2) return sqr(s);

    res.f_x=power(s.f_x,n);

    if(slope::calc_slope)
      {
       res.f_c=power(s.f_c,n);

       if(even(n)) //n ist gerade
	 {	  
	   interval x_d=_interval(Inf(s.f_x));
	   interval c_d=_interval(Inf(s.f_c));
	   
	   interval x_u=_interval(Sup(s.f_x));
	   interval c_u=_interval(Sup(s.f_c));

           if( !(0.0<=(x_d-c_d)) && !(0.0<=(x_u-c_u)))
	     {
	       //konvex

	       interval fxd=power(_interval(Inf(s.f_x)),n);
	       interval fxu=power(_interval(Sup(s.f_x)),n);
 
	       interval fcd=power(_interval(Inf(s.f_c)),n);
	       interval fcu=power(_interval(Sup(s.f_c)),n);

	       real h_d=Inf((fxd-fcd)/(x_d-c_d));	      
	       real h_u=Sup((fxu-fcu)/(x_u-c_u));

	       interval sl=_interval(h_d,h_u);   
	       for(i=1; i<=s.dim; i++)
		 {
		   res.f_s[i]=s.f_s[i]*sl;
		 }	       
	       return res;
	     }

	   if( (0.0<=(x_d-c_d)) && !(0.0<=(x_u-c_u)))
	     {
	       //konvex
	       //linke Seite durch Ableitung

	       interval fxu=power(_interval(Sup(s.f_x)),n);
	       interval fcu=power(_interval(Sup(s.f_c)),n);
	       
	       real h_d=Inf(_interval(n)*power(x_d,n-1));	      
	       real h_u=Sup((fxu-fcu)/(x_u-c_u));

	       interval sl=_interval(h_d,h_u);   
	       for(i=1; i<=s.dim; i++)
		 {
		   res.f_s[i]=s.f_s[i]*sl;
		 }	       
	       return res;
	     }
	   if( !(0.0<=(x_d-c_d)) && (0.0<=(x_u-c_u)))
	     {
	       //konvex 
	       //rechte Seite durch Ableitung

	       interval fxd=power(_interval(Inf(s.f_x)),n);
	       interval fcd=power(_interval(Inf(s.f_c)),n);

	       real h_d=Inf((fxd-fcd)/(x_d-c_d));	      
	       real h_u=Sup(_interval(n)*power(x_u,n-1));

	       interval sl=_interval(h_d,h_u);   
	       for(i=1; i<=s.dim; i++)
		 {
		   res.f_s[i]=s.f_s[i]*sl;
		 }	       
	       return res;
	     }	   
   
	 }//n even
       else //(n ist ungerade)
	 {
	   if(Inf(s.f_x)>0.0)
	       {
		 real h_d, h_u;

		 if(s_1_down(s.f_x, s.f_c, n, h_d, power)
		    && 
		    s_2_up(s.f_x, s.f_c, n, h_u, power))
		   {
		     //konvex
		     interval sl=_interval(h_d,h_u);   
		     for(i=1; i<=s.dim; i++)
		       {
			 res.f_s[i]=s.f_s[i]*sl;
		       }
		     return res;
		   }

		 if(!s_1_down(s.f_x, s.f_c, n, h_d, power)
		    && 
		    s_2_up(s.f_x, s.f_c, n, h_u, power))
		   {
		     //konvex 
		     //linke Seite durch Ableitung

		     interval x_d=_interval(Inf(s.f_x));
                     h_d=Inf(_interval(n)*power(x_d,n-1));

		     interval sl=_interval(h_d,h_u);   
		     for(i=1; i<=s.dim; i++)
		       {
			 res.f_s[i]=s.f_s[i]*sl;
		       }
		     return res;
		   }


		 if(s_1_down(s.f_x, s.f_c, n, h_d, power)
		    && 
		    !s_2_up(s.f_x, s.f_c, n, h_u, power))
		   {
		     //konvex 
		     //rechte Seite durch Ableitung

		     interval x_u=_interval(Sup(s.f_x));
		     h_u=Sup(_interval(n)*power(x_u,n-1));

		     interval sl=_interval(h_d,h_u);   
		     for(i=1; i<=s.dim; i++)
		       {
			 res.f_s[i]=s.f_s[i]*sl;
		       }
		     return res;
		   }		
 
	       }//Inf>0
	   if(Sup(s.f_x)<0.0)
	     {
		 real h_d, h_u;

		 if(s_1_up(s.f_x, s.f_c, n, h_u, power)
		    && 
		    s_2_down(s.f_x, s.f_c, n, h_d, power))
		   {
		     //konkav 

		     interval sl=_interval(h_d,h_u);   
		     for(i=1; i<=s.dim; i++)
		       {
			 res.f_s[i]=s.f_s[i]*sl;
		       }
		     return res;
		   }

		 if(s_1_up(s.f_x, s.f_c, n, h_u, power)
		    && 
		    !s_2_down(s.f_x, s.f_c, n, h_d, power))
		   {
		     //konkav 
		     //linke Seite ueber Ableitung

		     interval x_u=_interval(Sup(s.f_x));		  
                     h_d=Inf(_interval(n)*power(x_u,n-1));

		     interval sl=_interval(h_d,h_u);   
		     for(i=1; i<=s.dim; i++)
		       {
			 res.f_s[i]=s.f_s[i]*sl;
		       }
		     return res;
		   }

		 if(!s_1_up(s.f_x, s.f_c, n, h_u, power)
		    && 
		    s_2_down(s.f_x, s.f_c, n, h_d, power))
		   {
		     //konkav 
		     //rechte Seite ueber Ableitung

		     interval x_d=_interval(Inf(s.f_x));		  
                     h_u=Sup(_interval(n)*power(x_d,n-1));
   
		     interval sl=_interval(h_d,h_u);   
		     for(i=1; i<=s.dim; i++)
		       {
			 res.f_s[i]=s.f_s[i]*sl;
		       }
		     return res;
		   } 
	     }//Sup<0
	 }//n ungerade
	    
        //sonst Ableitung
       interval der=_interval(n)*power(s.f_x,n-1);
       for(i=1; i<=s.dim; i++)
	 {
	   res.f_s[i]=s.f_s[i]*der;
	 }
       return res;
      }
    return res;
  }//n>0
  else //n<0
    {
      res=_interval(1.0)/power(s,-n);
      return res;
    }
}

/*------------------------------------------------------------------------

pow(x,alpha) (=exp(alpha*ln(x)))

------------------------------------------------------------------------*/
slope pow(const slope& s, const interval& alpha)
{
 slope res(s.dim);
 res.f_x=pow(s.f_x,alpha);

 if(slope::calc_slope)
  {
   res=exp(alpha*ln(s)); //andere Darstellung fuer s hoch alpha
  }
 return res; 
}

/*------------------------------------------------------------------------

ln  konkav fuer alle x>0, streng monoton wachsend

------------------------------------------------------------------------*/
slope ln(const slope& s)
{
 slope res(s.dim);
 int i;
 res.f_x=ln(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=ln(s.f_c);
   interval h=s.f_x;

   real h_d, h_u;

   if(s_1_up(s.f_x, s.f_c, h_u, ln)
      && 
      s_2_down(s.f_x, s.f_c, h_d, ln))
     {
       //konkav 
       interval sl=_interval(h_d,h_u);
       for(i=1; i<=s.dim; i++)
	 {
	   res.f_s[i]=s.f_s[i]*sl;
	 }
       return res;
     }
   else
    {
      if( (!s_1_up(s.f_x, s.f_c, h_u, ln)) //==0
           && 
          s_2_down(s.f_x, s.f_c, h_d, ln) )
         {
	  //konkav 
         interval h1=_interval(Inf(s.f_x)); 
        
          if(0.0<=h1) //0 im Intervall
	   {
	     cout << "Fehler ln: 0 im Intervall" << endl; 
	     exit(1); //Funktion nicht auswertbar
	   }
	  else h_u=Sup(_interval(1.0)/h1);

	  interval sl=_interval(h_d,h_u);
	  for(i=1; i<=s.dim; i++)
	    {
	      res.f_s[i]=s.f_s[i]*sl;
	    }
	  return res;
        }
     else
      { 
	if( s_1_up(s.f_x, s.f_c, h_u, ln)
	     && 
	    (!s_2_down(s.f_x, s.f_c, h_d, ln)) ) //==0
	 {
	   //konkav 
          interval h2=_interval(Sup(s.f_x)); 

          if(0.0<=h2) //0 im Intervall
	   {
	     cout << "Fehler ln: 0 im Intervall" << endl; 
	     exit(1); //Funktion nicht auswertbar
	   }
	  else h_d=Inf(_interval(1.0)/h2);	  

	  interval sl=_interval(h_d,h_u);
	  for(i=1; i<=s.dim; i++)
	    {
	      res.f_s[i]=s.f_s[i]*sl;
	    }
	   return res;
	 }
       else
	{
	  if(0.0<=h) //0 im Intervall
	   {
	     cout << "Fehler ln: 0 im Intervall" << endl; 
	     exit(1); //Funktion nicht auswertbar
	   }
	 else 
	   {
	    for(i=1; i<=s.dim; i++)
	      { 
		res.f_s[i]=s.f_s[i]/h;
	      }
	   }
	}
      }
    }

  }
 return res; 
}

/*------------------------------------------------------------------------

exp  konvex fuer alle x aus R, streng monoton wachsend

------------------------------------------------------------------------*/
slope exp(const slope& s)
{
 slope res(s.dim);
 int i;
 res.f_x=exp(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=exp(s.f_c);

   real h_d, h_u;

   if(s_1_down(s.f_x, s.f_c, h_d, exp)
      && 
      s_2_up(s.f_x, s.f_c, h_u, exp))
     {
       //konvex 
       interval sl=_interval(h_d,h_u);
       for(i=1; i<=s.dim; i++)
	{
	  res.f_s[i]=s.f_s[i]*sl;
	}
       return res;
     }

   if( (!s_1_down(s.f_x, s.f_c, h_d, exp)) //==0
      && 
        s_2_up(s.f_x, s.f_c, h_u, exp) )
     {
       //konvex 
       h_d=Inf(res.f_x);

       interval sl=_interval(h_d,h_u);
       for(i=1; i<=s.dim; i++)
	{
	  res.f_s[i]=s.f_s[i]*sl;
	}
       return res;
     }

   if( s_1_down(s.f_x, s.f_c, h_d, exp)
      && 
       (!s_2_up(s.f_x, s.f_c, h_u, exp)) ) //==0
     {
       //konvex 
       h_u=Sup(res.f_x);

       interval sl=_interval(h_d,h_u);
       for(i=1; i<=s.dim; i++)
	{
	  res.f_s[i]=s.f_s[i]*sl;
	}
       return res;
     }
   //sonst Ableitung
     for(i=1; i<=s.dim; i++)
	{
	  res.f_s[i]=s.f_s[i]*res.f_x; 
	}
  }
 return res; 
}

/*------------------------------------------------------------------------

Slopes der trigonometrische Funktionen ueber Ableitungen berechnet

------------------------------------------------------------------------*/
slope sin(const slope& s)
{
 slope res(s.dim);
 res.f_x=sin(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=sin(s.f_c);
   interval der=cos(s.f_x);
   for(int i=1; i<=s.dim; i++)
     {
       res.f_s[i]=s.f_s[i]*der;
     }
  }
 return res; 
}

slope cos(const slope& s)
{
 slope res(s.dim);
 res.f_x=cos(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=cos(s.f_c);
   interval der=-sin(s.f_x);
   for(int i=1; i<=s.dim; i++)
     {
       res.f_s[i]=s.f_s[i]*der;
     }
  }
 return res; 
}

slope tan(const slope& s)
{
 slope res(s.dim);
 res.f_x=tan(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=tan(s.f_c);
   interval der=sqr(cos(s.f_x));
   for(int i=1; i<=s.dim; i++)
     {
       res.f_s[i]=s.f_s[i]/der;
     }
  }
 return res; 
}

slope cot(const slope& s)
{
 slope res(s.dim);
 res.f_x=cot(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=cot(s.f_c);
   interval der=-sqr(sin(s.f_x));
   for(int i=1; i<=s.dim; i++)
     {
       res.f_s[i]=s.f_s[i]/der;
     }
  }
 return res; 
}

/*------------------------------------------------------------------------

asin -1<=x<=0 konkav, 0<=x<=1 konvex, streng monoton steigend

------------------------------------------------------------------------*/
slope asin(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=asin(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=asin(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, asin)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, asin))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }   
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng wachsend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, asin)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, asin))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }       
	   return res;
	 }
     }  

    //ansonsten Ableitung
   interval der=sqrt(_interval(1.0)-sqr(s.f_x));
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res; 
}

/*------------------------------------------------------------------------

acos -1<=x<=0 konvex, 0<=x<=1 konkav, streng fallend

------------------------------------------------------------------------*/
slope acos(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=acos(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=acos(s.f_c);
   
   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng fallend
       
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acos)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acos))
	 { 
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }     
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng fallend
      real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, acos)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, acos))
	 {
 	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }     
	   return res;
	 } 
     } 
   
   //ansonsten Ableitung
   interval der=-sqrt(_interval(1.0)-sqr(s.f_x));
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res; 
}

/*------------------------------------------------------------------------

atan x<=0 konvex, 0<=x konkav, streng monoton steigend

------------------------------------------------------------------------*/
slope atan(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=atan(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=atan(s.f_c);
   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng wachsend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, atan)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, atan))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }    
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;
       if(s_1_down(s.f_x, s.f_c, h_d, atan)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, atan))
	 { 
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }  
     } 
   
   //ansonsten Ableitung
   interval der=_interval(1.0)+sqr(s.f_x);
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res; 
}

/*------------------------------------------------------------------------

acot x<=0 konkav, 0<=x konvex, streng momoton fallend

------------------------------------------------------------------------*/
slope acot(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=acot(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=acot(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng fallend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, acot)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, acot))
	 { 
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }    
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng fallend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acot)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acot))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }       
     } 

   //ansonsten Ableitung
   interval der=-(_interval(1.0)+sqr(s.f_x));
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res; 
}

/*------------------------------------------------------------------------

sinh x<=0 konkav, 0<=x konvex , streng monoton wachsend

------------------------------------------------------------------------*/
slope sinh(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=sinh(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=sinh(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng wachsend
      real h_d, h_u;

      if(s_1_down(s.f_x, s.f_c, h_d, sinh)
	 && 
	 s_2_up(s.f_x, s.f_c, h_u, sinh))
	{
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	  return res;
	}   
     }
	   
   if(Sup(s.f_x)<0.0)
     {
      //konkav u. streng wachsend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, sinh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, sinh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }
     } 
  
   //ansonsten Ableitung
   interval der=cosh(s.f_x);
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]*der;
  }
 return res;  

}

/*------------------------------------------------------------------------

cosh konvex in ganz R

------------------------------------------------------------------------*/
slope cosh(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=cosh(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=cosh(s.f_c);

   interval fxd=cosh(_interval(Inf(s.f_x)));
   interval fxu=cosh(_interval(Sup(s.f_x)));
 
   interval fcd=cosh(_interval(Inf(s.f_c)));
   interval fcu=cosh(_interval(Sup(s.f_c)));

   interval x_d=_interval(Inf(s.f_x));
   interval c_d=_interval(Inf(s.f_c));
	   
   interval x_u=_interval(Sup(s.f_x));
   interval c_u=_interval(Sup(s.f_c));

   if(!(0.0<=(x_d-c_d)) && !(0.0<=(x_u-c_u)))
     {
      //konvex 
      real h_d=Inf((fxd-fcd)/(x_d-c_d));	      
      real h_u=Sup((fxu-fcu)/(x_u-c_u));

      interval sl=_interval(h_d,h_u);
      for(i=1; i<=s.dim; i++)
       {
	res.f_s[i]=s.f_s[i]*sl;
       }
      return res;
     }

   //sonst Ableitung
   interval der=sinh(s.f_x);
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]*der;   
  }
 return res;
}

/*------------------------------------------------------------------------

tanh x<=0 konvex, 0<=x konkav , streng monoton wachsend

------------------------------------------------------------------------*/
slope tanh(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=tanh(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=tanh(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng wachsend
        real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, tanh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, tanh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }    
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, tanh)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, tanh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }
     }  
   //ansonsten Ableitung
   interval der=sqr(cosh(s.f_x));
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res;  
}

/*------------------------------------------------------------------------

coth x<0 konkav, 0<x konvex , streng monoton fallend

------------------------------------------------------------------------*/

slope coth(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=coth(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=coth(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng fallend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, coth)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, coth))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }         
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng fallend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, coth)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, coth))
	 { 
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }
     }  
   //ansonsten Ableitung
   interval der=-sqr(sinh(s.f_x));
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res;  
}

/*------------------------------------------------------------------------

asinh x<=0 konvex, 0<=x konkav , streng monoton wachsend

------------------------------------------------------------------------*/
slope asinh(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=asinh(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=asinh(s.f_c);
   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng wachsend
        real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, asinh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, asinh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     } 
	   return res;
	 }           
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, asinh)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, asinh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }      
     } 
   
   //ansonsten Ableitung
   interval der=sqrt(_interval(1.0)+sqr(s.f_x));
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res; 
}

/*------------------------------------------------------------------------

acosh(x) konkav fuer alle x>=1, streng monoton wachsend

------------------------------------------------------------------------*/
slope acosh(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=acosh(s.f_x);
 if(slope::calc_slope)
  {
    res.f_c=acosh(s.f_c);
    if(Inf(s.f_c)>1.0)
     {
       //konkav u. streng wachsend
      real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acosh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acosh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }    
     }

    interval h=s.f_x;
    if(1.0<=h) //1 im Intervall
      {
       cout << "Fehler acosh: 1 im Intervall" << endl; 
       exit(1); //Funktion nicht auswertbar
      }
     else 
       {
	 interval der= sqrt(sqr(s.f_x)-_interval(1.0));
         for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
       }
  }
 return res;
}

/*------------------------------------------------------------------------

atanh(x) x>=0: konvex ,x<=0: konkav, streng monoton wachsend

------------------------------------------------------------------------*/
slope atanh(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=atanh(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=atanh(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng wachsend
      real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, atanh)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, atanh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }      
	   return res;
	 }  
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng wachsend
      real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, atanh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, atanh))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }
     }  
    //ansonsten Ableitung
    interval der=_interval(1.0)-sqr(s.f_x);
    for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res;
}

/*------------------------------------------------------------------------

acoth x<0 konkav, 0<x konvex , streng monoton fallend

------------------------------------------------------------------------*/
slope acoth(const slope& s)
{
 slope res(s.dim);
 int i;
 res.f_x=acoth(s.f_x);

 if(slope::calc_slope)
  {
   res.f_c=acoth(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng fallend
      real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, acoth)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, acoth))
	 {
           interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng fallend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acoth)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acoth))
	 {
	   interval sl=_interval(h_d,h_u);
	   for(i=1; i<=s.dim; i++)
	     {
	       res.f_s[i]=s.f_s[i]*sl;
	     }
	   return res;
	 }
     }  
   //ansonsten Ableitung
   interval der=_interval(1.0)-sqr(s.f_x);
   for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]/der;
  }
 return res;  
}


/*------------------------------------------------------------------------

abs : nichtdiff. Funktion, vgl. Kearfott: Interval Extensions of Non-Smooth
      Functions for Global Optimization and Nonlinear System Solvers,
      Computing 57, p.154, Formel 13

------------------------------------------------------------------------*/
slope abs(const slope& s)
{
 slope res(s.dim);
 int i;

 res.f_x=abs(s.f_x);
 if(slope::calc_slope)
  {
    res.f_c=abs(s.f_c);
    if(Sup(s.f_x)<0)
      {
	for(i=1; i<=s.dim; i++)  res.f_s[i]=-s.f_s[i];
        return res;
      }
    if(Inf(s.f_x)>0) 
      {
	for(i=1; i<=s.dim; i++)	res.f_s[i]=s.f_s[i];
	return res;
      }
    //sonst
    interval help=h_abs(Inf(s.f_x),s.f_c)|h_abs(Sup(s.f_x),s.f_c);
    for(i=1; i<=s.dim; i++) res.f_s[i]=help*s.f_s[i];
  }
 return res;
}

/*------------------------------------------------------------------------

max : nichtdiff. Funktion, vgl. Kearfott: Interval Extensions of Non-Smooth
      Functions for Global Optimization and Nonlinear System Solvers,
      Computing 57, p.155, Formel 18

------------------------------------------------------------------------*/
slope max(const slope& s, const slope& t)
{

  if(s.dim!=t.dim)
    {cout << "Fehler in max(slope& , slope& )" << endl; exit(1);}

 slope res(s.dim);
 int i;

 res.f_x=h_max(s.f_x, t.f_x);


 if(slope::calc_slope)
  {
    res.f_c=h_max(s.f_c,t.f_c);

    if(Inf(s.f_x)>Sup(t.f_x)) 
      {
	for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i];
	return res;
      }
    if(Sup(s.f_x)<Inf(t.f_x)) 
      {
	for(i=1; i<=s.dim; i++) res.f_s[i]=t.f_s[i];
	return res;
      }
    //sonst
    for(i=1; i<=s.dim; i++) res.f_s[i]=s.f_s[i]|t.f_s[i];
  }
return res;
}

/*------------------------------------------------------------------------

min: auf max zurueckfuehren

------------------------------------------------------------------------*/
slope min(const slope& s, const slope& t)
{
 slope res, s_max, t_max;
 s_max=-s;
 t_max=-t;
 res=-max(s_max, t_max);
 return res;
}

/*------------------------------------------------------------------------

chi : nichtdiff. Funktion, vgl. Kearfott: Interval Extensions of Non-Smooth
      Functions for Global Optimization and Nonlinear System Solvers,
      Computing 57, p.152, Formeln 3 u.7

------------------------------------------------------------------------*/
slope chi(const slope& s, const slope& q, const slope& r)
{
 if(q.dim!=r.dim)
  {cout << "Fehler in chi(slope& , slope& )" << endl; exit(1);}
 slope res(r.dim);
 int i;

 res.f_x=chi(s.f_x, q.f_x, r.f_x);
 if(slope::calc_slope)
  {
    res.f_c=chi(s.f_c, q.f_c, r.f_c);
    for(i=1; i<=s.dim; i++) res.f_s[i]=chi(s.f_x, q.f_s[i], r.f_s[i]); 
  }
 return res;
}



///////////////////////////////////////////////////////////////
//
//Implementierung der Klasse slope_vector
//
///////////////////////////////////////////////////////////////

slope_vector::slope_vector()
{
 dim=1;
 lb=1;
 ub=1;
 dim_slopes=1; 
 sv=new slope[dim](dim_slopes);  
}

//------------------------------------------------------------------

slope_vector::slope_vector(int Lb, int Ub, int dimension_slope=1)
{
 dim=Ub-Lb+1;
 lb=Lb;
 ub=Ub;
 dim_slopes=dimension_slope; 
 sv=new slope[dim](dim_slopes);  
}

//------------------------------------------------------------------



slope_vector::slope_vector(const slope_vector& s)
{
 dim=s.dim;
 lb=s.lb;
 ub=s.ub;
 dim_slopes=s.dim_slopes;
 sv=new slope[dim](dim_slopes); 
 for(int i=0; i<dim; i++) sv[i]=s.sv[i]; 
}

slope_vector::~slope_vector()
{
 delete[] sv;
}

//------------------------------------------------------------------


slope_vector& slope_vector::operator=(const slope_vector& s)
{
  if (this!=&s)
     {
      delete[] sv;
      dim=s.dim;
      lb=s.lb;
      ub=s.ub;
      dim_slopes=s.dim_slopes;
      sv=new slope[dim](dim_slopes);
      for(int i=0; i<dim; i++) sv[i]=s.sv[i];
     }
   return *this;
}

//------------------------------------------------------------------


slope& slope_vector::operator[](int n)
{
 if(n>ub)
   {
    cout << "Fehler in slope_vector::operator[]: zu grosser Index" << endl;
    exit(1);
   }
 if(n<lb)
   {
    cout << "Fehler in slope_vector::operator[]: zu kleiner Index" << endl;
    exit(1);
   }   
 return sv[n-lb]; //Index verschoben!!
}

slope& slope_vector::operator[](int n) const
{
 if(n>ub)
   {
    cout << "Fehler in slope_vector::operator[]: zu grosser Index" << endl;
    exit(1);
   }
 if(n<lb)
   {
    cout << "Fehler in slope_vector::operator[]: zu kleiner Index" << endl;
    exit(1);
   }   
 return sv[n-lb]; //Index verschoben!!
}
//------------------------------------------------------------------

//Grenzen zurueckgeben
int Lb(const slope_vector& x)
{
 return x.lb;
}

int Ub(const slope_vector& x)
{
 return x.ub;
}

//------------------------------------------------------------------


//Initialisierungsfunktionen
slope_vector InitVar(const ivector& x)
{
 int dimension=Ub(x)-Lb(x)+1;

 slope_vector erg(Lb(x),Ub(x),dimension);

 int i;
 for(i=erg.lb; i<=erg.ub; i++) 
   {
 
    erg[i]=var_slope(dimension, i-erg.lb+1, x[i]);
   }
 return erg;
}

//------------------------------------------------------------------

slope_vector InitVar(const rvector& x)
{
 int dimension=Ub(x)-Lb(x)+1;

 slope_vector erg(Lb(x),Ub(x),dimension);
 
 int i;
 for(i=erg.lb; i<=erg.ub; i++) 
   {
    erg[i]=var_slope(dimension, i-erg.lb+1, x[i]);
   }
 return erg;
}

//------------------------------------------------------------------

slope_vector InitVar(const ivector& x,const  ivector& c)
{
 int dimension=Ub(x)-Lb(x)+1;
 int dimension_c=Ub(c)-Lb(c)+1;

 if(dimension!=dimension_c)
    {cout << "Fehler in InitVar(ivector& , ivector& )" << endl; exit(1);};

 if(Lb(x)!=Lb(c))
    {cout << "Fehler in InitVar(ivector& , ivector& ): Lb(x)!=Lb(c)" << endl; exit(1);};

 slope_vector erg(Lb(x),Ub(x),dimension);
 
 int i;
 for(i=erg.lb; i<=erg.ub; i++) 
   {
    erg[i]=var_slope(dimension, i-erg.lb+1, x[i], c[i]);
   }
 return erg;
}

//------------------------------------------------------------------

slope_vector InitVar(const ivector& x,const  rvector& c)
{
 int dimension=Ub(x)-Lb(x)+1;
 int dimension_c=Ub(c)-Lb(c)+1;

 if(dimension!=dimension_c)
    {cout << "Fehler in InitVar(ivector& , rvector& )" << endl; exit(1);};

 if(Lb(x)!=Lb(c))
    {cout << "Fehler in InitVar(ivector& , rvector& ): Lb(x)!=Lb(c)" << endl;
     exit(1);};

 slope_vector erg(Lb(x),Ub(x),dimension);
 
 int i;
 for(i=erg.lb; i<=erg.ub; i++) 
   {
    erg[i]=var_slope(dimension, i-erg.lb+1, x[i], c[i]);
   }
 return erg;
}

//------------------------------------------------------------------


//Entwicklungspkt.-e als Mittelpkt.-e waehlen
slope_vector Init_mid_Var(const ivector& x)
{
 int dimension=Ub(x)-Lb(x)+1;

 slope_vector erg(Lb(x),Ub(x),dimension);

 int i;
 for(i=erg.lb; i<=erg.ub; i++) 
   {
    erg[i]=init_mid_var(dimension, i-erg.lb+1, x[i]);
   }

 return erg;

}

//------------------------------------------------------------------


//Ausgabe
void print_slope_vector( const slope_vector& s)
{
 for(int i=Lb(s); i<=Ub(s); i++) print_slope(s[i]);
}

//////////////////////////////////////////////////////////////////////////
//
//einige vorgefertigte Prozeduren
//
//////////////////////////////////////////////////////////////////////////


/*------------------------------------------------------------------------

Prozedur berechnet nur Funktionswert fuer Intervallvektor x

------------------------------------------------------------------------*/
void feval(slope (*f)(const slope_vector& ), const ivector& x , interval& f_value)
{
 slope::unset_calc_slope();
 slope_vector x_in=Init_mid_Var(x); 
 slope erg;
 erg=f(x_in);
 f_value=get_fx(erg);
 slope::set_calc_slope();
}

/*------------------------------------------------------------------------

Prozedur berechnet Funktionswert und Steigungvektor fuer  x

------------------------------------------------------------------------*/
void fseval(slope (*f)(const slope_vector& ), const ivector& x, 
	    interval& f_value, ivector& steigungen)
{
 slope::set_calc_slope();
 slope_vector x_in=Init_mid_Var(x); 
 slope erg;
 erg=f(x_in);
 f_value=get_fx(erg);
 steigungen=get_fs(erg);
}

} //namespace cxsc
