/*
**  SLOPE_FOR: C-XSC classes for one- and multidimensional (interval)
**             slope arithmetic in forward mode
**
**  Copyright (C) 1999-2001 M. Braeuer, W. Hofschuster, W. Kraemer
**                          Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


//////////////////////////////////////////////////////////////
//
//Implementierung  slope.cpp
//
//////////////////////////////////////////////////////////////

#include "slope.h"

namespace cxsc {

///////////////////////////////////////////////////////////////
//
// Klasse slope_type
//
///////////////////////////////////////////////////////////////

//statische Komponenten
int slope_type::calc_slope_type=1; 

//----------------------------------------------------------------------

void slope_type::set_calc_slope_type()
{
 slope_type::calc_slope_type=1;
}

//----------------------------------------------------------------------

void slope_type::unset_calc_slope_type()
{
 slope_type::calc_slope_type=0;
}

//----------------------------------------------------------------------

int slope_type::get_calc_slope_type()
{
 return slope_type::calc_slope_type;
}

//----------------------------------------------------------------------

//Konstruktoren (fuer Unabhaengige)
slope_type::slope_type(const real& x)
{
 f_x=x;
 f_c=x;
 f_s=_interval(1.0);   
}

slope_type::slope_type(const interval& x)
{
 f_x=x;
 f_c=x;
 f_s=_interval(1.0);
}

slope_type::slope_type(const interval& x, const interval& c)
{
 f_x=x;
 f_c=c;
 f_s=_interval(1.0);
}

slope_type::slope_type(const interval& x,const  real& c)
{
 f_x=x;
 f_c=_interval(c);
 f_s=_interval(1.0);
}

//----------------------------------------------------------------------

//Initialisierungsfunktion fuer Unabhaengige
slope_type var_slope_type(const interval& c1, const interval& c2)
{
 slope_type res;
 res.f_x=c1;
 res.f_c=c2;
 res.f_s=_interval(1.0); 
 return res;
}

//----------------------------------------------------------------------


//Initialisierung der Konstanten
slope_type const_slope_type(const real& c)
{
 slope_type res;
 res.f_x=_interval(c);
 res.f_c=_interval(c);
 res.f_s=_interval(0.0); 
 return res;
}

//----------------------------------------------------------------------

slope_type const_slope_type(const interval& c)
{
 slope_type res;
 res.f_x=c;
 res.f_c=c;
 res.f_s=_interval(0.0); 
 return res;
}

//----------------------------------------------------------------------

slope_type const_slope_type(const interval& c1, const interval& c2)
{
 slope_type res;
 res.f_x=c1;
 res.f_c=c2;
 res.f_s=_interval(0.0); 
 return res;
}

//----------------------------------------------------------------------

slope_type const_slope_type(const interval& c1, const real& c2)
{
 slope_type res;
 res.f_x=c1;
 res.f_c=_interval(c2);
 res.f_s=_interval(0.0);
 return res; 
}

//----------------------------------------------------------------------

//Initialisierung der 2.Intervallkomponente mit Mittelpunkt
//bei einer Abhaengigen 
slope_type init_mid_var(const interval& x)
{
 slope_type res;
 res.f_x=x;
 res.f_c=_interval(mid(x));
 res.f_s=_interval(1.0);
 return res; 
}

//----------------------------------------------------------------------

//Initialisierung der 2.Intervallkomponente mit Mittelpunkt
//bei einer Konstanten
slope_type init_mid_const(const interval& x)
{
 slope_type res;
 res.f_x=x;
 res.f_c=_interval(mid(x));
 res.f_s=_interval(0.0);
 return res; 
}

//----------------------------------------------------------------------

//folgende Funktionen geben private Elemente zurueck
interval get_fx(const slope_type& s)
{
 return s.f_x;
}

interval get_fc(const slope_type& s)
{
 return s.f_c;
}

interval get_fs(const slope_type& s)
{
 return s.f_s;
}

//----------------------------------------------------------------------

//Ausgabe eines Objekts vom Typ slope_type
void print_slope_type(const slope_type& s)
{
  cout << SaveOpt << SetPrecision(16,14)<< "f_x " << s.f_x << endl;
  cout << "f_c " << s.f_c << endl;
  cout << "f_s " << s.f_s << endl << RestoreOpt << endl;; 
}

//----------------------------------------------------------------------

//Ueberladen der Operatoren

//unaerer Operator -
slope_type operator-(const slope_type& s)
{
 slope_type res;
 res.f_x=-s.f_x;
 if(slope_type::calc_slope_type)
   {
    res.f_c=-s.f_c;
    res.f_s=-s.f_s;
   }
 return res;
}

//----------------------------------------------------------------------

//binaere Operatoren (slope_type,slope_type)
slope_type operator-(const slope_type& s, const slope_type& t)
{
 slope_type res;
 res.f_x=s.f_x-t.f_x;
 if(slope_type::calc_slope_type)
   {
    res.f_c=s.f_c-t.f_c;
    res.f_s=s.f_s-t.f_s;
   }
 return res;
}

//----------------------------------------------------------------------

slope_type operator+(const slope_type& s, const slope_type& t)
{
 slope_type res;
 res.f_x=s.f_x+t.f_x;
 if(slope_type::calc_slope_type)
   {
    res.f_c=s.f_c+t.f_c;
    res.f_s=s.f_s+t.f_s;
   }
 return res;
}

//----------------------------------------------------------------------

slope_type operator*(const slope_type& s, const slope_type& t)
{
 slope_type res;
 res.f_x=s.f_x*t.f_x;
 if(slope_type::calc_slope_type)
   {
    res.f_c=s.f_c*t.f_c;
    res.f_s=s.f_s*t.f_x+s.f_c*t.f_s; //Symetrie ausnutzen
    res.f_s=res.f_s & (s.f_x*t.f_s+s.f_s*t.f_c); //Durchschnitt
   }
 return res;
}

//----------------------------------------------------------------------

slope_type operator/(const slope_type& s, const slope_type& t)
{
 slope_type res;
 if(_interval(0.0)<=t.f_x) 
   {
     cout << "Fehler in / 0 in f_x" << endl; exit(1);
   }
 res.f_x=s.f_x/t.f_x;
 if(slope_type::calc_slope_type)
   {
     if(_interval(0.0)<=t.f_c) 
       {
	 cout << "Fehler in / 0 in f_c" << endl; exit(1);
       }
    res.f_c=s.f_c/t.f_c;

    res.f_s=(s.f_s-res.f_c*t.f_s)/t.f_x;  //Symetrie ausnutzen
    res.f_s=res.f_s & (s.f_s-res.f_x*t.f_s)/t.f_c; //Durchschnitt
   }
 return res;
}

//----------------------------------------------------------------------

//binaere Operatoren (slope_type,interval)

slope_type operator-(const slope_type& s, const interval& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s-t_slope;
 return res;
}

//----------------------------------------------------------------------

slope_type operator+(const slope_type& s, const interval& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s+t_slope;
 return res;
}

//----------------------------------------------------------------------

slope_type operator*(const slope_type& s, const interval& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s*t_slope;
 return res;
}

//----------------------------------------------------------------------

slope_type operator/(const slope_type& s, const interval& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s/t_slope;
 return res;
}

//----------------------------------------------------------------------

//binaere Operatoren (interval, slope_type)
slope_type operator-(const interval& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope-t;
 return res;
}

//----------------------------------------------------------------------

slope_type operator+(const interval& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope+t;
 return res;
}

//----------------------------------------------------------------------

slope_type operator*(const interval& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope*t;
 return res;
}

//----------------------------------------------------------------------

slope_type operator/(const interval& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope/t;
 return res;
} 

//----------------------------------------------------------------------

//binaere Operatoren (slope_type, real)
slope_type operator-(const slope_type& s, const real& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s-t_slope;
 return res;
}

//----------------------------------------------------------------------

slope_type operator+(const slope_type& s, const real& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s+t_slope;
 return res;
}

//----------------------------------------------------------------------

slope_type operator*(const slope_type& s, const real& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s*t_slope;
 return res;
}

//----------------------------------------------------------------------

slope_type operator/(const slope_type& s, const real& t)
{
 slope_type res;
 slope_type t_slope=const_slope_type(t);
 res=s/t_slope;
 return res;
}

//----------------------------------------------------------------------

//binaere Operatoren (real, slope_type)
slope_type operator-(const real& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope-t;
 return res;
}

//----------------------------------------------------------------------

slope_type operator+(const real& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope+t;
 return res;
}

//----------------------------------------------------------------------

slope_type operator*(const real& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope*t;
 return res;
}

//----------------------------------------------------------------------

slope_type operator/(const real& s, const slope_type& t)
{
 slope_type res;
 slope_type s_slope=const_slope_type(s);
 res=s_slope/t;
 return res;
}

//----------------------------------------------------------------------

//Ueberladen der Standardfunktionen

/*------------------------------------------------------------------------

sqr(x) konvex fuer alle x aus R

------------------------------------------------------------------------*/
slope_type sqr(const slope_type& s)
{
 slope_type res;
 res.f_x=sqr(s.f_x);
 if(slope_type::calc_slope_type)
  {
    res.f_c=sqr(s.f_c);
    res.f_s=(s.f_x+s.f_c)*s.f_s;
  }
 return res;
}

/*------------------------------------------------------------------------

sqrt(x) konkav fuer alle x>=0, streng monoton wachsend

------------------------------------------------------------------------*/
slope_type sqrt(const slope_type& s)
{
 slope_type res;
 res.f_x=sqrt(s.f_x);
 if(slope_type::calc_slope_type)
  {
    res.f_c=sqrt(s.f_c);
    interval h=res.f_x+res.f_c;
    if(0.0<=h) //0 im Intervall
      {
       cout << "Fehler sqrt: 0 im Intervall" << endl; 
       exit(1); //Funktion nicht auswertbar
      }
     else res.f_s=s.f_s/h;
  }
 return res;
}

/*------------------------------------------------------------------------

sqrt(x,n ) konkav fuer alle x>=0 und n>0 aus N, streng monoton wachsend

------------------------------------------------------------------------*/
slope_type sqrt(const slope_type& s,int n)
{
 slope_type res;

 if(n==0) 
  {
    cout << "Fehler in sqrt(x,n): n=0" << endl;
    exit(1);
  }

 if(n>0)
   {
     if(n==1) return s;
     if(n==2) return sqrt(s);

     //n>=3:
     res.f_x=sqrt(s.f_x, n);

     if(slope_type::calc_slope_type)
       {
	 res.f_c=sqrt(s.f_c, n);

	 interval h=s.f_x;

         real h_d, h_u;

	 interval _x=_interval(Inf(s.f_x));
	 interval _c=_interval(Inf(s.f_c));

	 interval h1= _x-_c; 

	 if(!(_interval(0.0)<=h1)) 
	   {     
	     interval _fx=sqrt(_x,n);
	     interval _fc=sqrt(_c,n);
	     interval i_res=(_fx-_fc)/h1;
	     h_u = Sup(i_res); //s_1_up
	 
	     interval x_=_interval(Sup(s.f_x));
	     interval c_=_interval(Sup(s.f_c));

	     interval h2= x_-c_; 

	     if(!(_interval(0.0)<=h2))
	       {
		 interval fx_=sqrt(x_,n);
		 interval fc_=sqrt(c_,n);
		 interval i_res=(fx_-fc_)/h2;
		 h_d = Inf(i_res); //s_2_down

		 res.f_s=s.f_s*_interval(h_d,h_u);
		 return res;
	       }          
	   }
	 else
	   if(0.0<=h) //0 im Intervall
	     {
	       cout << "Fehler sqrt(x,n): 0 im Intervall" << endl; 
	       exit(1); //Funktion nicht auswertbar
	     }
	   else res.f_s=s.f_s*sqrt(h,n)/(_interval(n)*h);
       }
     return res;
   }
 else {res=_interval(1.0)/sqrt(s,-n); return res;}
}

/*------------------------------------------------------------------------

power(x,n )
    n>0 :   n gerade: konvex fuer alle x aus R fallend, wachsend
            n ungerade: konvex fuer x>=0 str. monoton wachsend
	                konkav fuer x<=0 str. monoton wachsend

------------------------------------------------------------------------*/
slope_type power(const slope_type& s,int n)
{
 slope_type res;

 if(n==0) 
  {
    if(_interval(0.0)<=s.f_x)
     {
      cout << "Fehler in power(x,n): n=0 und 0 im Intervall" << endl;
      exit(1);
     }     
    real c=1.0;
    res=const_slope_type(c);
    return res;
  }

 if(n>0)
     {
      if(n==1) return s;
      if(n==2) return sqr(s);

      res.f_x=power(s.f_x,n);

      if(slope_type::calc_slope_type)
	{
	  res.f_c=power(s.f_c,n);

	  if(even(n)) //n ist gerade
	    { 
	      interval x_d=_interval(Inf(s.f_x));
	      interval c_d=_interval(Inf(s.f_c));
	   
	      interval x_u=_interval(Sup(s.f_x));
	      interval c_u=_interval(Sup(s.f_c));

	      if( !(0.0<=(x_d-c_d)) && !(0.0<=(x_u-c_u)))
		{
		  //konvex 

		  interval fxd=power(_interval(Inf(s.f_x)),n);
		  interval fxu=power(_interval(Sup(s.f_x)),n);
 
		  interval fcd=power(_interval(Inf(s.f_c)),n);
		  interval fcu=power(_interval(Sup(s.f_c)),n);

		  real h_d=Inf((fxd-fcd)/(x_d-c_d));	      
		  real h_u=Sup((fxu-fcu)/(x_u-c_u));
	       
		  res.f_s=s.f_s*_interval(h_d,h_u);
		  return res;
		}

	      if( (0.0<=(x_d-c_d)) && !(0.0<=(x_u-c_u)))
		{
		  //konvex 
		  //linke Seite durch Ableitung

		  interval fxu=power(_interval(Sup(s.f_x)),n);
		  interval fcu=power(_interval(Sup(s.f_c)),n);
	       
		  real h_d=Inf(_interval(n)*power(x_d,n-1));	      
		  real h_u=Sup((fxu-fcu)/(x_u-c_u));
	       
		  res.f_s=s.f_s*_interval(h_d,h_u);
		  return res;
		}
	      if( !(0.0<=(x_d-c_d)) && (0.0<=(x_u-c_u)))
		{
		  //konvex 
		  //rechte Seite durch Ableitung

		  interval fxd=power(_interval(Inf(s.f_x)),n);
		  interval fcd=power(_interval(Inf(s.f_c)),n);

		  real h_d=Inf((fxd-fcd)/(x_d-c_d));	      
		  real h_u=Sup(_interval(n)*power(x_u,n-1));
	       
		  res.f_s=s.f_s*_interval(h_d,h_u);
		  return res;
		}	   
   
	    }//n even
	  else //(n ist ungerade)
	    {
	      if(Inf(s.f_x)>0.0)
	       {
		 real h_d, h_u;

		 if(s_1_down(s.f_x, s.f_c, n, h_d, power)
		    && 
		    s_2_up(s.f_x, s.f_c, n, h_u, power))
		   {
		     //konvex u. streng wachsend
		     res.f_s=s.f_s*_interval(h_d,h_u);
		     return res;
		   }


		 if(!s_1_down(s.f_x, s.f_c, n, h_d, power)
		    && 
		    s_2_up(s.f_x, s.f_c, n, h_u, power))
		   {
		     //konvex u. streng wachsend
		     //linke Seite durch Ableitung

		     interval x_d=_interval(Inf(s.f_x));
                     h_d=Inf(_interval(n)*power(x_d,n-1));
		     res.f_s=s.f_s*_interval(h_d,h_u);
		     return res;
		   }


		 if(s_1_down(s.f_x, s.f_c, n, h_d, power)
		    && 
		    !s_2_up(s.f_x, s.f_c, n, h_u, power))
		   {
		     //konvex u. streng wachsend
		     //rechte Seite durch Ableitung

		     interval x_u=_interval(Sup(s.f_x));
		     h_u=Sup(_interval(n)*power(x_u,n-1));
		     res.f_s=s.f_s*_interval(h_d,h_u);
		     return res;
		   }		
 
	       }
	      if(Sup(s.f_x)<0.0)
	       {
		 real h_d, h_u;

		 if(s_1_up(s.f_x, s.f_c, n, h_u, power)
		    && 
		    s_2_down(s.f_x, s.f_c, n, h_d, power))
		   {
		     //konkav u. streng wachsend
		     res.f_s=s.f_s*_interval(h_d,h_u);
		     return res;
		   }

		 if(s_1_up(s.f_x, s.f_c, n, h_u, power)
		    && 
		    !s_2_down(s.f_x, s.f_c, n, h_d, power))
		   {
		     //konkav u. streng wachsend
		     //linke Seite ueber Ableitung

		     interval x_u=_interval(Sup(s.f_x));		     
                     h_d=Inf(_interval(n)*power(x_u,n-1));
		     res.f_s=s.f_s*_interval(h_d,h_u);
		     return res;
		   }

		 if(!s_1_up(s.f_x, s.f_c, n, h_u, power)
		    && 
		    s_2_down(s.f_x, s.f_c, n, h_d, power))
		   {
		     //konkav u. streng wachsend
		     //rechte Seite ueber Ableitung

		     interval x_d=_interval(Inf(s.f_x));		     
                     h_u=Sup(_interval(n)*power(x_d,n-1));
		     res.f_s=s.f_s*_interval(h_d,h_u);
		     return res;
		   }
	       } // Sup(s.f_x)<0
	    }//else n odd

	  //sonst Ableitung
	  res.f_s=s.f_s*_interval(n)*power(s.f_x,n-1);
	  return res;
	}
      return res;
     }//n>0
      else //n<0
	{
	  res=_interval(1.0)/power(s,-n);
	  return res;
	}
}

/*------------------------------------------------------------------------

pow(x,alpha) (=exp(alpha*ln(x)))

------------------------------------------------------------------------*/
slope_type pow(const slope_type& s, const interval& alpha)
{
 slope_type res;
 res.f_x=pow(s.f_x,alpha);

 if(slope_type::calc_slope_type)
  {
   res=exp(alpha*ln(s)); //andere Darstellung fuer s hoch alpha
  }
 return res; 
}

/*------------------------------------------------------------------------

ln  konkav fuer alle x>0, streng monoton wachsend

------------------------------------------------------------------------*/
slope_type ln(const slope_type& s)
{
 slope_type res;
 res.f_x=ln(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=ln(s.f_c);
   interval h=s.f_x;

   real h_d, h_u;

   if(s_1_up(s.f_x, s.f_c, h_u, ln)
      && 
      s_2_down(s.f_x, s.f_c, h_d, ln))
     {
       //konkav u. streng wachsend
       res.f_s=s.f_s*_interval(h_d,h_u);
       return res;
     }
   else
    {
      if( (!s_1_up(s.f_x, s.f_c, h_u, ln)) //==0
           && 
          s_2_down(s.f_x, s.f_c, h_d, ln) )
         {
	  //konkav u. streng wachsend
          interval h1=_interval(Inf(s.f_x)); 
        
          if(0.0<=h1) //0 im Intervall
	   {
	     cout << "Fehler ln: 0 im Intervall" << endl; 
	     exit(1); //Funktion nicht auswertbar
	   }
	  else h_u=Sup(_interval(1.0)/h1);

		 //divu(1.0,Inf(s.f_x));

	  res.f_s=s.f_s*_interval(h_d,h_u);
	  return res;
        }
     else
      { 
	if( s_1_up(s.f_x, s.f_c, h_u, ln)
	     && 
	    (!s_2_down(s.f_x, s.f_c, h_d, ln)) ) //==0
	 {
	   //konkav u. streng wachsend
          interval h2=_interval(Sup(s.f_x)); 

          if(0.0<=h2) //0 im Intervall
	   {
	     cout << "Fehler ln: 0 im Intervall" << endl; 
	     exit(1); //Funktion nicht auswertbar
	   }
	  else h_d=Inf(_interval(1.0)/h2);

		 //divd(1.0,Sup(s.f_x));

	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
       else
	{
	  if(0.0<=h) //0 im Intervall
	   {
	     cout << "Fehler ln: 0 im Intervall" << endl; 
	     exit(1); //Funktion nicht auswertbar
	   }
	 else res.f_s=s.f_s/h;
	}
      }
    }

  }
 return res; 
}


/*------------------------------------------------------------------------

exp  konvex fuer alle x aus R, streng monoton wachsend

------------------------------------------------------------------------*/
slope_type exp(const slope_type& s)
{
 slope_type res;
 res.f_x=exp(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=exp(s.f_c);

   real h_d, h_u;

   if(s_1_down(s.f_x, s.f_c, h_d, exp)
      && 
      s_2_up(s.f_x, s.f_c, h_u ,exp))
     {
       //konvex u. streng wachsend
       res.f_s=s.f_s*_interval(h_d,h_u);
       return res;
     }

   if( (!s_1_down(s.f_x, s.f_c, h_d, exp)) //==0
      && 
        s_2_up(s.f_x, s.f_c, h_u, exp) )
     {
       //konvex u. streng wachsend
       h_d=Inf(res.f_x);
       res.f_s=s.f_s*_interval(h_d,h_u);
       return res;
     }

   if( s_1_down(s.f_x, s.f_c, h_d, exp)
      && 
       (!s_2_up(s.f_x, s.f_c,h_u, exp)) ) //==0
     {
       //konvex u. streng wachsend
       h_u=Sup(res.f_x);
       res.f_s=s.f_s*_interval(h_d,h_u);
       return res;
     }

     res.f_s=s.f_s*res.f_x;
  }
 return res; 
}

/*------------------------------------------------------------------------

Slopes der trigonometrische Funktionen ueber Ableitungen berechnet

------------------------------------------------------------------------*/
slope_type sin(const slope_type& s)
{
 slope_type res;
 res.f_x=sin(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=sin(s.f_c);
   res.f_s=s.f_s*cos(s.f_x);
  }
 return res; 
}

slope_type cos(const slope_type& s)
{
 slope_type res;
 res.f_x=cos(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=cos(s.f_c);
   res.f_s=-s.f_s*sin(s.f_x);
  }
 return res; 
}

slope_type tan(const slope_type& s)
{
 slope_type res;
 res.f_x=tan(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=tan(s.f_c);
   res.f_s=s.f_s/sqr(cos(s.f_x));
  }
 return res; 
}

slope_type cot(const slope_type& s)
{
 slope_type res;
 res.f_x=cot(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=cot(s.f_c);
   res.f_s=-s.f_s/sqr(sin(s.f_x));
  }
 return res; 
}

/*------------------------------------------------------------------------

asin -1<=x<=0 konkav, 0<=x<=1 konvex, streng monoton steigend

------------------------------------------------------------------------*/
slope_type asin(const slope_type& s)
{
 slope_type res;
 res.f_x=asin(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=asin(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, asin)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, asin))
	 {
      
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }   
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng wachsend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, asin)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, asin))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
     }  

    //ansonsten Ableitung
   res.f_s=s.f_s/sqrt(1.0-sqr(s.f_x));
  }
 return res; 
}


/*------------------------------------------------------------------------

acos -1<=x<=0 konvex, 0<=x<=1 konkav, streng fallend

------------------------------------------------------------------------*/
slope_type acos(const slope_type& s)
{
 slope_type res;
 res.f_x=acos(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=acos(s.f_c);
   
   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng fallend
       
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acos)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acos))
	 {     
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }     
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng fallend
      real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, acos)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, acos))
	 {      
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 } 
     } 
   
   //ansonsten Ableitung
   res.f_s=-s.f_s/sqrt(1.0-sqr(s.f_x));
  }
 return res; 
}

/*------------------------------------------------------------------------

atan x<=0 konvex, 0<=x konkav, streng monoton steigend

------------------------------------------------------------------------*/
slope_type atan(const slope_type& s)
{
 slope_type res;
 res.f_x=atan(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=atan(s.f_c);
   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng wachsend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, atan)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, atan))
	 {
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }    
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;
       if(s_1_down(s.f_x, s.f_c, h_d, atan)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, atan))
	 {    
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }  
     } 
   
   //ansonsten Ableitung
   res.f_s=s.f_s/(1.0+sqr(s.f_x));
  }
 return res; 
}

/*------------------------------------------------------------------------

acot x<=0 konkav, 0<=x konvex, streng momoton fallend

------------------------------------------------------------------------*/
slope_type acot(const slope_type& s)
{
 slope_type res;
 res.f_x=acot(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=acot(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng fallend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, acot)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, acot))
	 {      
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }    
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng fallend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acot)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acot))
	 {     
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }       
     } 

   //ansonsten Ableitung
   res.f_s=-s.f_s/(1.0+sqr(s.f_x));
  }
 return res; 
}

/*------------------------------------------------------------------------

sinh x<=0 konkav, 0<=x konvex , streng monoton wachsend

------------------------------------------------------------------------*/
slope_type sinh(const slope_type& s)
{
 slope_type res;
 res.f_x=sinh(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=sinh(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng wachsend
      real h_d, h_u;

      if(s_1_down(s.f_x, s.f_c, h_d, sinh)
	 && 
	 s_2_up(s.f_x, s.f_c, h_u, sinh))
	{
       
	  res.f_s=s.f_s*_interval(h_d,h_u);
	  return res;
	}   
     }
	   
   if(Sup(s.f_x)<0.0)
     {
      //konkav u. streng wachsend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, sinh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, sinh))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
     } 
  
   //ansonsten Ableitung
   res.f_s=s.f_s*cosh(s.f_x);
  }
 return res;  

}

/*------------------------------------------------------------------------

cosh konvex in ganz R

------------------------------------------------------------------------*/
slope_type cosh(const slope_type& s)
{
 slope_type res;
 res.f_x=cosh(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=cosh(s.f_c);

   interval fxd=cosh(_interval(Inf(s.f_x)));
   interval fxu=cosh(_interval(Sup(s.f_x)));
 
   interval fcd=cosh(_interval(Inf(s.f_c)));
   interval fcu=cosh(_interval(Sup(s.f_c)));

   interval x_d=_interval(Inf(s.f_x));
   interval c_d=_interval(Inf(s.f_c));
	   
   interval x_u=_interval(Sup(s.f_x));
   interval c_u=_interval(Sup(s.f_c));

   if(!(0.0<=(x_d-c_d)) && !(0.0<=(x_u-c_u)))
     {
      //konvex 
      real h_d=Inf((fxd-fcd)/(x_d-c_d));	      
      real h_u=Sup((fxu-fcu)/(x_u-c_u));
	       
      res.f_s=s.f_s*_interval(h_d,h_u);
      return res;
     }

   //sonst Ableitung
   res.f_s=s.f_s*sinh(s.f_x);   
  }
 return res;
}

/*------------------------------------------------------------------------

tanh x<=0 konvex, 0<=x konkav , streng monoton wachsend

------------------------------------------------------------------------*/
slope_type tanh(const slope_type& s)
{
 slope_type res;
 res.f_x=tanh(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=tanh(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng wachsend
        real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, tanh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, tanh))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }    
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, tanh)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, tanh))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
     }  
   //ansonsten Ableitung
   res.f_s=s.f_s/sqr(cosh(s.f_x));
  }
 return res;  
}

/*------------------------------------------------------------------------

coth x<0 konkav, 0<x konvex , streng monoton fallend

------------------------------------------------------------------------*/

slope_type coth(const slope_type& s)
{
 slope_type res;
 res.f_x=coth(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=coth(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng fallend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, coth)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, coth))
	 {      
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }         
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng fallend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, coth)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, coth))
	 {     
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
     }  
   //ansonsten Ableitung
   res.f_s=-s.f_s/sqr(sinh(s.f_x));
  }
 return res;  
}

/*------------------------------------------------------------------------

asinh x<=0 konvex, 0<=x konkav , streng monoton wachsend

------------------------------------------------------------------------*/
slope_type asinh(const slope_type& s)
{
 slope_type res;
 res.f_x=asinh(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=asinh(s.f_c);
   if(Inf(s.f_x)>0.0)
     {
       //konkav u. streng wachsend
        real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, asinh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, asinh))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }           
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konvex u. streng wachsend
       real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, asinh)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, asinh))
	 {
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }      
     } 
   
   //ansonsten Ableitung
   res.f_s=s.f_s/sqrt(1.0+sqr(s.f_x));
  }
 return res; 
}

/*------------------------------------------------------------------------

acosh(x) konkav fuer alle x>=1, streng monoton wachsend

------------------------------------------------------------------------*/
slope_type acosh(const slope_type& s)
{
 slope_type res;
 res.f_x=acosh(s.f_x);
 if(slope_type::calc_slope_type)
  {
    res.f_c=acosh(s.f_c);
    if(Inf(s.f_c)>1.0)  //Entwicklungspunkt !=1, moeglicher Overflow
     {
       //konkav u. streng wachsend
      real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acosh )
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acosh))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }    
     }

    interval h=s.f_x;
    if(1.0<=h) //1 im Intervall
      {
       cout << "Fehler acosh: 1 im Intervall" << endl; 
       exit(1); //Funktion nicht auswertbar
      }
     else res.f_s=s.f_s/sqrt(sqr(s.f_x)-1.0);
  }
 return res;
}

/*------------------------------------------------------------------------

atanh(x) x>=0: konvex ,x<=0: konkav, streng monoton wachsend

------------------------------------------------------------------------*/
slope_type atanh(const slope_type& s)
{
 slope_type res;
 res.f_x=atanh(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=atanh(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng wachsend
      real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, atanh)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, atanh))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }  
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng wachsend
      real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, atanh)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, atanh))
	 {
       
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
     }  
    //ansonsten Ableitung
    res.f_s=s.f_s/(1.0-sqr(s.f_x));
  }
 return res;
}

/*------------------------------------------------------------------------

acoth x<0 konkav, 0<x konvex , streng monoton fallend

------------------------------------------------------------------------*/
slope_type acoth(const slope_type& s)
{
 slope_type res;
 res.f_x=acoth(s.f_x);

 if(slope_type::calc_slope_type)
  {
   res.f_c=acoth(s.f_c);

   if(Inf(s.f_x)>0.0)
     {
       //konvex u. streng fallend
      real h_d, h_u;

       if(s_1_down(s.f_x, s.f_c, h_d, acoth)
	  && 
	  s_2_up(s.f_x, s.f_c, h_u, acoth))
	 {      
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
     }
	   
   if(Sup(s.f_x)<0.0)
     {
       //konkav u. streng fallend
       real h_d, h_u;

       if(s_1_up(s.f_x, s.f_c, h_u, acoth)
	  && 
	  s_2_down(s.f_x, s.f_c, h_d, acoth))
	 {     
	   res.f_s=s.f_s*_interval(h_d,h_u);
	   return res;
	 }
     }  
   //ansonsten Ableitung
   res.f_s=s.f_s/(1.0-sqr(s.f_x));
  }
 return res;  
}


/*------------------------------------------------------------------------

abs : nichtdiff. Funktion, vgl. Kearfott: Interval Extensions of Non-Smooth
      Functions for Global Optimization and Nonlinear System Solvers,
      Computing 57, p.154, Formel 13

------------------------------------------------------------------------*/
slope_type abs(const slope_type& s)
{
 slope_type res;
 res.f_x=abs(s.f_x);
 if(slope_type::calc_slope_type)
  {
    res.f_c=abs(s.f_c);
    if(Sup(s.f_x)<0) {res.f_s=-s.f_s;return res;}
    if(Inf(s.f_x)>0) {res.f_s=s.f_s;return res;}
    res.f_s=(h_abs(Inf(s.f_x),s.f_c)|h_abs(Sup(s.f_x),s.f_c))*s.f_s;
  }
 return res;
}

/*------------------------------------------------------------------------

max : nichtdiff. Funktion, vgl. Kearfott: Interval Extensions of Non-Smooth
      Functions for Global Optimization and Nonlinear System Solvers,
      Computing 57, p.155, Formel 18

------------------------------------------------------------------------*/
slope_type max(const slope_type& s, const slope_type& t)
{
 slope_type res;
 res.f_x=h_max(s.f_x, t.f_x);
 if(slope_type::calc_slope_type)
  {
    res.f_c=h_max(s.f_c,t.f_c);

    if(Inf(s.f_x)>Sup(t.f_x)) {res.f_s=s.f_s;return res;}
    if(Sup(s.f_x)<Inf(t.f_x)) {res.f_s=t.f_s;return res;}
    res.f_s=s.f_s|t.f_s;
  }
return res;
}

/*------------------------------------------------------------------------

min: auf max zurueckfuehren

------------------------------------------------------------------------*/
slope_type min(const slope_type& s, const slope_type& t)
{
 slope_type res, s_max, t_max;
 s_max=-s;
 t_max=-t;
 res=-max(s_max, t_max);
 return res;
}

/*------------------------------------------------------------------------

chi : nichtdiff. Funktion, vgl. Kearfott: Interval Extensions of Non-Smooth
      Functions for Global Optimization and Nonlinear System Solvers,
      Computing 57, p.152, Formeln 3 u.7

------------------------------------------------------------------------*/
slope_type chi(const slope_type& s, const slope_type& q, const slope_type& r)
{
 slope_type res;
 res.f_x=chi(s.f_x, q.f_x, r.f_x);
 if(slope_type::calc_slope_type)
  {
    res.f_c=chi(s.f_c, q.f_c, r.f_c);
    res.f_s=chi(s.f_x, q.f_s, r.f_s); 
  }
 return res;
}

//////////////////////////////////////////////////////////////////////////
//
//einige vorgefertigte Prozeduren
//
//////////////////////////////////////////////////////////////////////////


/*------------------------------------------------------------------------

Prozedur berechnet nur Funktionswert im Intervall x

------------------------------------------------------------------------*/
void feval(slope_type (*f)( const slope_type& ), const interval& x,  interval& f_value)
{
 slope_type::unset_calc_slope_type();
 slope_type s(x);
 slope_type res;
 res=f(s);
 f_value=get_fx(res);
 slope_type::set_calc_slope_type(); 
}


/*------------------------------------------------------------------------

Prozedur berechnet Funktionswert und Steigung im Intervall x

------------------------------------------------------------------------*/
void fseval(slope_type (*f)( const slope_type& ), const interval& x, 
	    interval& f_value,
	    interval& steigung)
{
 slope_type::set_calc_slope_type();
 slope_type s=init_mid_var(x);
 slope_type res;
 res=f(s);

 f_value=get_fx(res);
 steigung=get_fs(res);

}

} //namespace cxsc
